''
''
''     QuadRotor - Full Stability
'' -- Jason Dorie --               
''               
''

{
  TODO:
        Use integrated gyro value for P, current gyro measurement for D
        Use accelerometer to bias integrated gyro result
                Store last 128 results in HUB RAM, keep a running sum, SHR 7 for current value

  DONE:
        Temperature drift compensation
        Convert ITG-3200 code to PASM
        Add ADXL345 readings to ITG code
        Integration of gyro values
}


CON
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000


OBJ
  Gyro  : "9DOF-pasm.spin"                              ' 1 cog
  Flt   : "Float32Full-Limited_Routines-1_Cog"          ' 1 cog                        
  Dbg   : "FullDuplexSerial.spin"                       ' 1 cog

  'Main loop                                            ' 1 cog

  DCM   : "PropDCM28.spin"                              ' 0 cogs
                          

VAR
  long counter
  long LastTime, CurTime

  long dGx, dGy, dGz                                    ' Working variables used by the DCM code
  long dAx, dAy, dAz

  long XAngle, YAngle, ZAngle



PUB Main | v, x , y
 
  Dbg.Start( 31, 30, 0, 115200 )
  Gyro.Start( 20, 21 )
  Flt.Start

  RunDCM  'Can be put into its own cog if desired




PUB RunDCM | tx, tz, loopTime

  DCM.Init

  '{
  'Start by gathering an average accelerometer value to use as "down"
  'Assumes the board is flat / level on startup
  dAx := 0
  dAy := 0
  dAz := 0
  loopTime := cnt

  repeat counter from 1 to 256
    dAx += Gyro.GetAX
    dAy += Gyro.GetAY
    dAz += Gyro.GetAZ

    waitcnt( constant(80_000_000 / 200) + loopTime )
    loopTime += constant(80_000_000 / 200)

  dAx := dAx << (DCM#FixedBits - (8 + 8))               '8 bits of accel value, 8 bits of loop accumulation
  dAy := dAy << (DCM#FixedBits - (8 + 8))
  dAz := dAz << (DCM#FixedBits - (8 + 8))
  

  'BiasCorrect SHOULD NOT include 1G gravity
  DCM.SetBiasCorrect( dAx, dAy, dAz - DCM#Fixed_One )
  '}


  loopTime := cnt

  repeat
    dGx := Gyro.GetRX
    dGy := Gyro.GetRY
    dGz := Gyro.GetRZ
    dAx := Gyro.GetAX
    dAy := Gyro.GetAY
    dAz := Gyro.GetAZ


    DCM.SetValues( dGx, dGy, dGz, dAx, dAy, dAz )

    'tx := cnt
    DCM.Update
    'tx := cnt - tx
    'dbg.dec( tx )
    'dbg.tx( 1 )    

    'Get the YAW vector in 2D, convert the direction of the user stick into a 2D vector as well
    'dot product will tell me how close and in which direction I am

    'Get the M[1,0] and M[1,2] components of the body matrix (Y components of the X and Z vectors)
    'Those are the SIN of the angle between the ground and the craft


    'It's worth noting that the range returned from ASIN is +/-(PI/2), or +/- 90 degrees
    'therefore the ASIN results will only be useful if the craft is within 89 degrees of "upright"
    'unless additional work is done to determine full range angular values.



    'Uncomment this block when you just want to see the rotation values
    {
    tx := Flt.ASin( Flt.FFloatDiv28( DCM.GetMatrixValue(5) ) )     'Convert to int, then divide by 1<<28
    tz := Flt.ASin( Flt.FFloatDiv28( DCM.GetMatrixValue(3) ) )     'Convert to int, then divide by 1<<28 

    XAngle := Flt.FRound( Flt.FMul( tx , constant(65536.0 / (PI / 2.0)) ) ) 
    ZAngle := Flt.FRound( Flt.FMul( tz , constant(-65536.0 / (PI / 2.0)) ) )

    'For heading, I want an actual angular value, so this returns me an int between 0 & 65535, where 0 is forward
    'Shifting it up 2 to allow for more bits of precision when accumulating the desired yaw value
    YAngle := Flt.FRound( Flt.FMul( Flt.Atan2( Flt.FFloat(DCM.GetMatrixValue(6)), Flt.FFloat(DCM.GetMatrixValue(8))), constant(32768.0 / PI) ) ) & 65535 

    dbg.tx(1)
    dbg.hex( XAngle, 4 )
    dbg.tx(32)
    dbg.hex( YAngle, 4 )
    dbg.tx(32)
    dbg.hex( ZAngle, 4 )
    dbg.tx(11)
    '}

    'Debug print the accelerometer values
    {
    dbg.tx(1)
    dbg.hex( dAx, 4 )
    dbg.tx(32)
    dbg.hex( dAy, 4 )
    dbg.tx(32)
    dbg.hex( dAz, 4 )
    dbg.tx(11)
    '}

    'Uncomment for the DCM Visualizer with LocalMode = true, or with AccelCalibrate = true
    {
    dbg.tx( $77 )
    dbg.tx( $77 )
    dbg.tx( dGx >> 8 )
    dbg.tx( dGx )
    dbg.tx( dGy >> 8 )
    dbg.tx( dGy )
    dbg.tx( dGz >> 8 )
    dbg.tx( dGz )
    dbg.tx( dAx >> 8 )
    dbg.tx( dAx )
    dbg.tx( dAy >> 8 )
    dbg.tx( dAy )
    dbg.tx( dAz >> 8 )
    dbg.tx( dAz )
    '}     

    'Uncomment for the DCM Visualizer with LocalMode = false
    '{
    dbg.tx( $78 )
    dbg.tx( $78 )

    dbg.tx( DCM.GetMatrixValue(0) >> 22 )
    dbg.tx( DCM.GetMatrixValue(0) >> 14 )
    dbg.tx( DCM.GetMatrixValue(1) >> 22 )
    dbg.tx( DCM.GetMatrixValue(1) >> 14 )
    dbg.tx( DCM.GetMatrixValue(2) >> 22 )
    dbg.tx( DCM.GetMatrixValue(2) >> 14 )

    dbg.tx( DCM.GetMatrixValue(3) >> 22 )
    dbg.tx( DCM.GetMatrixValue(3) >> 14 )
    dbg.tx( DCM.GetMatrixValue(4) >> 22 )
    dbg.tx( DCM.GetMatrixValue(4) >> 14 )
    dbg.tx( DCM.GetMatrixValue(5) >> 22 )
    dbg.tx( DCM.GetMatrixValue(5) >> 14 )

    'The last row is now computed in the vis tool
    '}

    waitcnt( constant(80_000_000 / 200) + loopTime )
    loopTime += constant(80_000_000 / 200)

     